#!/usr/bin/env python
"""
 Juju add-credential validation using aws credential yaml file and to
 verify credentials specified in the yaml file.

 Usage: python assess_add_credentials.py --credential-file=/path/to/yaml_file
"""
import yaml
import csv
import sys
import logging
from argparse import ArgumentParser
from utility import (
    add_basic_testing_arguments,
    JujuAssertionError,
    temp_yaml_file,
    )

from deploy_stack import BootstrapManager
from jujupy import (
    client_from_config,
    )
from credential_utility import (
    write_aws_config_file,
    begin_test,
    client_credentials_to_details,
)

log = logging.getLogger("assess_add_credentials")


def make_aws_credential_dict(cloud_name, access_key, secret_key):
    """Returns aws credentials dictionary data structure

    :param cloud_name: cloud name for aws
    :param access_key: access-key specified in credential file
    :param secret_key: secret-key specefied in credential file
    :return: aws credential dictionary

    """
    return {
        'credentials': {
            'aws': {
                cloud_name: {
                    'auth-type': 'access-key',
                    'access-key': access_key,
                    'secret-key': secret_key,
                }
            }
        }
    }


def verify_aws_add_credentials(
      client, cloud_name, expected_access_key, expected_secret_key):
    """Verify that 'cloud_name', credentials are the expected provided values.

    :param client: juju client
    :param cloud_name: Name of the cloud specified in credential file
    :param expected_access_key: Expected access-key
    :param expected_secret_key: Expected secret key

    """
    actual_credentials_details = yaml.safe_load(
      client.list_credentials('aws', show_secret=True))
    actual_credentials = actual_credentials_details[
      'credentials']['aws']['cloud-credentials'][cloud_name]['details']
    if expected_access_key != actual_credentials['access-key']:
        raise JujuAssertionError('AWS access-key mismatch!.')
    if expected_secret_key != actual_credentials['secret-key']:
        raise JujuAssertionError('AWS secret-key mismatch!.')
    log.info("AWS credential validated successfully")


def make_aws_cloud_data(credential_file, cloud_name):
    """Generate aws credentials dict from credential file

    :param credential_file : Expected to be a filepath string.

    """
    with open(credential_file, 'rt') as acf:
        access_key, secret_key = get_aws_credentials_from_file(acf)
    aws_cloud_data = make_aws_credential_dict(
        cloud_name, access_key, secret_key)
    return aws_cloud_data, access_key, secret_key


def assess_aws_credentials(client_base, credential_file):
    """Perform juju add-credential using yaml file and verify it.

    :param client_base: juju client
    :param credential_file: credential_file is expected to be a filepath string

    """
    cloud_name = 'juju'
    cloud_details = make_aws_cloud_data(credential_file, cloud_name)
    cloud_data, access_key, secret_key = cloud_details
    with temp_yaml_file(cloud_data) as aws_cloud:
        with begin_test(client_base) as (client, tmp_scratch_dir):
            client.add_credential('aws', aws_cloud)
            verify_aws_add_credentials(
                client, cloud_name, access_key, secret_key)


def add_and_bootstrap(bs_manager, upload_tools):
    """Ensure we can bootstrap after adding credentials."""
    cloud_name = 'juju'
    with begin_test(bs_manager.client) as (client_na, tmp_scratch_dir):
        bs_manager.client.env.juju_home = client_na.env.juju_home
        bs_manager.tear_down_client.env.juju_home = client_na.env.juju_home
        real_credentials = prepare_credential_file(
            bs_manager.client, cloud_name, tmp_scratch_dir)
        # Reset cached credentials before adding them via 'juju add-credential'
        bs_manager.client.env.credentials = {}
        with bs_manager.top_context() as machines:
            with bs_manager.bootstrap_context(
                    machines,
                    omit_config=bs_manager.client.bootstrap_replaces):
                aws_cloud_data = make_aws_credential_dict(
                    cloud_name, real_credentials['access_key'],
                    real_credentials['secret_key'])
                with temp_yaml_file(aws_cloud_data) as aws_cloud:
                    bs_manager.client.add_credential('aws', aws_cloud)
                    bs_manager.client.env.load_yaml()
                    bs_manager.client.bootstrap(
                        upload_tools=upload_tools,
                        bootstrap_series=bs_manager.series)
                    bs_manager.client.kill_controller()


def prepare_credential_file(client, cloud_name, tmp_dir):
    """Create temp aws Credential file

    :Param: client: Juju Client
    :Param: cloud_name: cloud name
    :Param: tmp_dir: String path of temp directory

    """
    real_credentials = client_credentials_to_details(client)
    if client.env.provider == 'ec2':
        write_aws_config_file(
            cloud_name, tmp_dir, real_credentials['access_key'],
            real_credentials['secret_key'])
    return real_credentials


def get_aws_credentials_from_file(aws_csv_file):
    """Read credential values from aws credential csv file.

    :param: aws_csv_file: aws_csv_file is expected to be a file-like object.

    """
    credential_details = {}
    reader = csv.reader(aws_csv_file, delimiter='=', quotechar='|')
    for row in reader:
        credential_details[row[0]] = row[1]
    return (credential_details['AWSAccessKeyId'],
            credential_details['AWSSecretKey'])


def parse_args(argv):
    """Parse all arguments."""
    parser = ArgumentParser(
        description="Testing add-credential command.")

    add_basic_testing_arguments(parser)

    parser.add_argument(
       '--credential-file', required=True, action='store',
       help='AWS credential file to be used during bootstrap'
       )
    return parser.parse_args(argv)


def main(argv=None):
    args = parse_args(argv)
    client = client_from_config(args.env, args.juju_bin, False)
    assess_aws_credentials(client, args.credential_file)
    bs_manager = BootstrapManager.from_args(args)
    add_and_bootstrap(bs_manager, args.upload_tools)
    return 0


if __name__ == '__main__':
    sys.exit(main())
