"""Tests for assess_add_credentials module."""

import assess_add_credentials as aac
from assess_add_credentials import (
  make_aws_credential_dict,
  parse_args,
  get_aws_credentials_from_file,
  verify_aws_add_credentials,
  make_aws_cloud_data,
  assess_aws_credentials,
  add_and_bootstrap,
  prepare_credential_file,
  )
from jujupy import fake_juju_client
from tests import (
    TestCase,
    )

from mock import (
    Mock,
    patch,
    )
from utility import (
    JujuAssertionError,
    temp_dir,
    )
from deploy_stack import BootstrapManager
import StringIO
import tempfile
from textwrap import dedent
import yaml


class TestParseArgs(TestCase):

    def test_common_args(self):
        credential_file = './rootkey.csv'
        args = parse_args(["an-env", "/bin/juju", "/tmp/logs",
                           "an-env-mod", '--credential-file', credential_file])
        self.assertEqual("an-env", args.env)
        self.assertEqual("/bin/juju", args.juju_bin)
        self.assertEqual("/tmp/logs", args.logs)
        self.assertEqual("an-env-mod", args.temp_env_name)
        self.assertEqual(False, args.debug)
        self.assertEqual(credential_file, args.credential_file)


class TestawsCredentials(TestCase):

    def test_make_aws_cloud_data(self):
        exp_access_key = 'KeyId123'
        exp_secret_key = 'SecretKey321'
        cloud_name = 'juju'
        exp_cloud_data = {
            'credentials': {
                'aws': {
                    cloud_name: {
                        'auth-type': 'access-key',
                        'access-key': exp_access_key,
                        'secret-key': exp_secret_key,
                    }
                }
            }
        }
        with tempfile.NamedTemporaryFile() as fp:
            fp.write(dedent("""\
                   AWSAccessKeyId=KeyId123
                   AWSSecretKey=SecretKey321
                     """))
            fp.flush()

            actual_output = make_aws_cloud_data(fp.name, cloud_name)
            cloud_data, access_key, secret_key = actual_output
            actual_cloud_data = cloud_data
            actual_access_key = access_key
            actual_secret_key = secret_key
            self.assertEqual(actual_cloud_data, exp_cloud_data)
            self.assertEqual(actual_access_key, exp_access_key)
            self.assertEqual(actual_secret_key, exp_secret_key)

    def test_assess_aws_credential(self):
        mock_begin_client = Mock()
        mock_client = Mock()
        with tempfile.NamedTemporaryFile() as fp:
            fp.write(dedent("""\
                    AWSAccessKeyId=KeyId123
                    AWSSecretKey=SecretKey321
                      """))
            fp.flush()
            with patch.object(
                    aac, "begin_autoload_test", autospec=True) as m_bat:
                m_bat.return_value.__enter__.return_value = (
                    mock_begin_client, 'tmp_dir_path')
                with patch.object(
                        aac, 'verify_aws_add_credentials',
                        autospec=True) as vaac:
                    assess_aws_credentials(mock_client, fp.name)
                    vaac.assert_called_once_with(
                        mock_begin_client, 'juju', 'KeyId123', 'SecretKey321')

    def test_make_aws_credential_dict_returns_dict_with_details(self):
        cloud_name = 'cloud_name'
        access_key = 'access_key'
        secret_key = 'secret_key'
        actual_output = make_aws_credential_dict(
          cloud_name, access_key, secret_key)
        expected_output = {
         'credentials': {
            'aws': {
                cloud_name: {
                    'auth-type': 'access-key',
                    'access-key': access_key,
                    'secret-key': secret_key,
                  }
              }
           }
        }
        self.assertDictEqual(actual_output, expected_output)

    def test_get_aws_credentials_from_file(self):
        expected_aws_access_key = 'TestAccessKey'
        expected_aws_secret_key = 'TestSecretKey'
        fake_csv_file_obj = StringIO.StringIO(
                     dedent("""\
                 AWSAccessKeyId=TestAccessKey
                 AWSSecretKey=TestSecretKey
                 """))
        actual_aws_access_key, actual_aws_secret_key = \
            get_aws_credentials_from_file(fake_csv_file_obj)
        self.assertEqual(actual_aws_access_key, expected_aws_access_key)
        self.assertEqual(actual_aws_secret_key, expected_aws_secret_key)

    def test_verify_aws_add_credentials_with_correct_values(self):
        cloud_name = 'cloud_name'
        access_key = 'test_access_key'
        secret_key = 'test_secret_key'
        actual_output = yaml.safe_dump({
            "credentials": {
                "aws": {
                    "cloud-credentials": {
                        cloud_name: {
                            "auth-type": "access-key", "details": {
                                "access-key": "{}".format(access_key),
                                "secret-key": "{}".format(secret_key)
                            }
                        }
                    }
                }
            }
        })
        mock_client = Mock()
        with patch.object(mock_client, "list_credentials",
                          return_value=actual_output, autospec=True):
            verify_aws_add_credentials(
                mock_client, cloud_name, access_key, secret_key)
        self.assertIn("AWS credential validated successfully",
                      self.log_stream.getvalue())

    def test_verify_aws_add_credentials_with_access_key_mismatch(self):
        cloud_name = 'cloud_name'
        actual_access_key = 'test_access_key'
        actual_secret_key = 'test_secret_key'
        expected_access_key = 'access_key'
        expected_secret_key = 'test_secret_key'
        actual_output = yaml.safe_dump({
            "credentials": {
                "aws": {
                    "cloud-credentials": {
                        cloud_name: {
                            "auth-type": "access-key", "details": {
                                "access-key": "{}".format(actual_access_key),
                                "secret-key": "{}".format(actual_secret_key)
                            }
                        }
                    }
                }
            }
        })
        mock_client = Mock()
        with patch.object(mock_client, "list_credentials",
                          return_value=actual_output, autospec=True):
            with self.assertRaisesRegexp(JujuAssertionError,
                                         'AWS access-key mismatch!.'):
                verify_aws_add_credentials(
                    mock_client, cloud_name, expected_access_key,
                    expected_secret_key)

    def test_verify_aws_add_credentials_with_secret_key_mismatch(self):
        cloud_name = 'cloud_name'
        actual_access_key = 'test_access_key'
        actual_secret_key = 'test_secret_key'
        expected_access_key = 'test_access_key'
        expected_secret_key = 'secret_key'
        actual_output = yaml.safe_dump({
            "credentials": {
                "aws": {
                    "cloud-credentials": {
                        cloud_name: {
                            "auth-type": "access-key", "details": {
                                "access-key": "{}".format(actual_access_key),
                                "secret-key": "{}".format(actual_secret_key)
                            }
                        }
                    }
                }
            }
        })
        mock_client = Mock()
        with patch.object(mock_client, "list_credentials",
                          return_value=actual_output, autospec=True):
            with self.assertRaisesRegexp(JujuAssertionError,
                                         'AWS secret-key mismatch!.'):
                verify_aws_add_credentials(
                    mock_client, cloud_name, expected_access_key,
                    expected_secret_key)

    def test_prepare_credential_file(self):
        expected_credential = {'access_key': 'accesskey123',
                               'secret_key': 'Secretkey321'}
        client = fake_juju_client()
        client.env._cloud_name = 'aws'
        client.env.credentials = {"credentials": {
                               "aws": {
                                "access-key": "accesskey123",
                                "secret-key": "Secretkey321"
                               }
        }}
        cloud_name = 'juju'
        with temp_dir() as tmp_dir:
            actual_credential = prepare_credential_file(
                client, cloud_name, tmp_dir)
            self.assertDictEqual(expected_credential, actual_credential)


class TestAddAndBootstrap(TestCase):
    def test_add_and_bootstrap(self):
        client = fake_juju_client()
        upload_tools = False
        real_credential_details = {
                        'access-key': 'keyid123',
                        'secret-key': 'Secretkey123',
                                   }
        credential_file_details = {'credentials': {'cloud': {
            'credentials': real_credential_details
        }}}

        def credential_check(*args, **kwargs):
            self.assertEqual(client.env.credentials, credential_file_details)
        with temp_dir() as log_dir:
            bs_manager = BootstrapManager(
                'env', client, client, None, [], None, None, None, None,
                log_dir, False, True, True)
            with patch.object(bs_manager.client, 'add_credential',
                              autospec=True, side_effect=credential_check)\
                    as run_add_mock:
                with patch.object(
                        bs_manager.client, 'bootstrap', autospec=True,
                        side_effect=credential_check) as bootstrap_mock:
                    add_and_bootstrap(bs_manager, upload_tools)
                    run_add_mock.assert_called_once_with(
                        client, real_credential_details, None)
                    bootstrap_mock.assert_called_once_with(
                        False, bootstrap_series=None)
